"""Constants for the Gruenbeck Cloud library."""

from datetime import timedelta
import re
from typing import Any, Final

import aiohttp
from yarl._url import DEFAULT_PORTS

# User Agent configuration
USER_AGENT_APP: Final = "Gruenbeck/354 CFNetwork/1209 Darwin/20.2.0"
USER_AGENT_WS: Final = (
    "Mozilla/5.0 (iPhone; CPU iPhone OS 14_2 like Mac OS X)"
    " AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148"
)

# API Parameters mapping
PARAM_NAME_CODE_CHALLENGE: Final = "code_challenge"
PARAM_NAME_CSRF_TOKEN: Final = "csrf_token"
PARAM_NAME_TENANT: Final = "tenant"
PARAM_NAME_TRANS_ID: Final = "transId"
PARAM_NAME_POLICY: Final = "policy"
PARAM_NAME_USERNAME: Final = "signInName"
PARAM_NAME_PASSWORD: Final = "password"
PARAM_NAME_CODE: Final = "code"
PARAM_NAME_CODE_VERIFIER: Final = "code_verifier"
PARAM_NAME_ACCESS_TOKEN: Final = "access_token"
PARAM_NAME_REFRESH_TOKEN: Final = "refresh_token"
PARAM_NAME_CONNECTION_ID: Final = "connection_id"
PARAM_NAME_DEVICE_ID: Final = "device_id"
PARAM_NAME_ENDPOINT: Final = "endpoint"


# Details needed for login
LOGIN_SCHEME: Final = "https"
LOGIN_HOST: Final = "gruenbeckb2c.b2clogin.com"
LOGIN_CODE_CHALLENGE_CHARS: Final = (
    "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
)
LOGIN_REFRESH_TIME_BEFORE_EXPIRE = timedelta(minutes=10)

# HTTP API Details
API_SCHEME: Final = "https"
API_HOST: Final = "prod-eu-gruenbeck-api.azurewebsites.net"
API_VERSION: Final = "2024-05-02"
API_GET_MG_INFOS_ENDPOINT: Final = ""  # Endpoint is empty for normal MG Infos
API_GET_MG_INFOS_ENDPOINT_PARAMETERS: Final = "parameters"
API_GET_MG_INFOS_ENDPOINT_SALT_MEASUREMENTS: Final = "measurements/salt"
API_GET_MG_INFOS_ENDPOINT_WATER_MEASUREMENTS: Final = "measurements/water"

# WS API Details
API_WS_SCHEME_HTTP: Final = "https"
API_WS_SCHEME_WS: Final = "wss"
API_WS_HOST: Final = "prod-eu-gruenbeck-signalr.service.signalr.net"
API_WS_REQUEST_TIMEOUT: Final = 2 * 60  # 2 minutes
API_WS_CLIENT_URL: Final = "/client/"
API_WS_CLIENT_QUERY: dict[str, str] = {
    "hub": "gruenbeck",
    "id": f"{{{PARAM_NAME_CONNECTION_ID}}}",
    "access_token": f"{{{PARAM_NAME_ACCESS_TOKEN}}}",
}
API_WS_CLIENT_HEADER: dict[str, str] = {
    "Upgrade": "websocket",
    "Host": "prod-eu-gruenbeck-signalr.service.signalr.net",
    "Origin": "null",
    "Pragma": "no-cache",
    "Cache-Control": "no-cache",
    "User-Agent": USER_AGENT_WS,
}
# There is a "%1E = Record Separator" char at the end of the string!
API_WS_INITIAL_MESSAGE: Final = '{"protocol":"json","version":1}'
API_WS_RESPONSE_TYPE_PING: Final = 6
API_WS_RESPONSE_TYPE_DATA: Final = 1
API_WS_RESPONSE_TYPE_DATA_TARGETS: list[str] = [
    "SendOneTimeMessageToDevice",
    "SendMessageToDevice",
]

# Web Request details
WEB_REQUESTS: dict[str, dict[str, Any]] = {
    "login_step_1": {
        "scheme": LOGIN_SCHEME,
        "host": LOGIN_HOST,
        "port": DEFAULT_PORTS.get(LOGIN_SCHEME),
        "path": (
            "/a50d35c1-202f-4da7-aa87-76e51a3098c6/b2c_1a_signinup/oauth2/v2.0/authorize"  # noqa: E501
        ),
        "method": aiohttp.hdrs.METH_GET,
        "use_cookies": True,
        "data": {},
        "json_data": False,
        "query_params": {
            "x-client-Ver": "0.8.0",
            "state": "NjkyQjZBQTgtQkM1My00ODBDLTn3MkYtOTZCQ0QyQkQ2NEE5",
            "client_info": "1",
            "response_type": "code",
            "code_challenge_method": "S256",
            "x-app-name": "Grünbeck",
            "x-client-OS": "14.3",
            "x-app-ver": "1.2.1",
            "scope": (
                "https://gruenbeckb2c.onmicrosoft.com/iot/user_impersonation openid profile offline_access"  # noqa: E501
            ),
            "x-client-SKU": "MSAL.iOS",
            "code_challenge": f"{{{PARAM_NAME_CODE_CHALLENGE}}}",
            "x-client-CPU": "64",
            "client-request-id": "F2929DED-2C9D-49F5-A0F4-31215427667C",
            "redirect_uri": "msal5a83cc16-ffb1-42e9-9859-9fbf07f36df8://auth",
            "client_id": "5a83cc16-ffb1-42e9-9859-9fbf07f36df8",
            "haschrome": "1",
            "return-client-request-id": "true",
            "x-client-DM": "iPhone",
        },
        "headers": {
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            "Accept-Encoding": "br, gzip, deflate",
            "Connection": "keep-alive",
            "Accept-Language": "de-de",
            "User-Agent": USER_AGENT_APP,
        },
    },
    "login_step_2": {
        "scheme": LOGIN_SCHEME,
        "host": LOGIN_HOST,
        "port": DEFAULT_PORTS.get(LOGIN_SCHEME),
        "path": f"{{{PARAM_NAME_TENANT}}}/SelfAsserted",
        "method": aiohttp.hdrs.METH_POST,
        "use_cookies": True,
        "data": {
            "request_type": "RESPONSE",
            "signInName": f"{{{PARAM_NAME_USERNAME}}}",
            "password": f"{{{PARAM_NAME_PASSWORD}}}",
        },
        "json_data": False,
        "query_params": {
            "tx": f"{{{PARAM_NAME_TRANS_ID}}}",
            "p": f"{{{PARAM_NAME_POLICY}}}",
        },
        "headers": {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "X-CSRF-TOKEN": f"{{{PARAM_NAME_CSRF_TOKEN}}}",
            "Accept": "application/json, text/javascript, */*; q=0.01",
            "X-Requested-With": "XMLHttpRequest",
            "Origin": "https://gruenbeckb2c.b2clogin.com",
            "User-Agent": USER_AGENT_APP,
        },
    },
    "login_step_3": {
        "scheme": LOGIN_SCHEME,
        "host": LOGIN_HOST,
        "port": DEFAULT_PORTS.get(LOGIN_SCHEME),
        "path": f"{{{PARAM_NAME_TENANT}}}/api/CombinedSigninAndSignup/confirmed",
        "method": aiohttp.hdrs.METH_GET,
        "use_cookies": True,
        "data": {},
        "json_data": False,
        "query_params": {
            "csrf_token": f"{{{PARAM_NAME_CSRF_TOKEN}}}",
            "tx": f"{{{PARAM_NAME_TRANS_ID}}}",
            "p": f"{{{PARAM_NAME_POLICY}}}",
        },
        "headers": {
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            "Accept-Encoding": "br, gzip, deflate",
            "Connection": "keep-alive",
            "Accept-Language": "de-de",
            "User-Agent": USER_AGENT_APP,
        },
    },
    "login_step_4": {
        "scheme": LOGIN_SCHEME,
        "host": LOGIN_HOST,
        "port": DEFAULT_PORTS.get(LOGIN_SCHEME),
        "path": f"{{{PARAM_NAME_TENANT}}}/oauth2/v2.0/token",
        "method": aiohttp.hdrs.METH_POST,
        "use_cookies": True,
        "data": {
            "client_info": "1",
            "scope": (
                "https://gruenbeckb2c.onmicrosoft.com/iot/user_impersonation openid profile offline_access"  # noqa: E501
            ),
            "code": f"{{{PARAM_NAME_CODE}}}",
            "grant_type": "authorization_code",
            "code_verifier": f"{{{PARAM_NAME_CODE_VERIFIER}}}",
            "redirect_uri": "msal5a83cc16-ffb1-42e9-9859-9fbf07f36df8://auth",
            "client_id": "5a83cc16-ffb1-42e9-9859-9fbf07f36df8",
        },
        "json_data": False,
        "query_params": {},
        "headers": {
            "Host": "gruenbeckb2c.b2clogin.com",
            "x-client-SKU": "MSAL.iOS",
            "Accept": "application/json",
            "x-client-OS": "14.3",
            "x-app-name": "Grünbeck",
            "x-client-CPU": "64",
            "x-app-ver": "1.2.0",
            "Accept-Language": "de-de",
            "client-request-id": "F2929DED-2C9D-49F5-A0F4-31215427667C",
            "x-ms-PkeyAuth": "1.0",
            "x-client-Ver": "0.8.0",
            "x-client-DM": "iPhone",
            "User-Agent": USER_AGENT_APP,
            "return-client-request-id": "true",
        },
    },
    "web_token_refresh": {
        "scheme": LOGIN_SCHEME,
        "host": LOGIN_HOST,
        "port": DEFAULT_PORTS.get(LOGIN_SCHEME),
        "path": f"{{{PARAM_NAME_TENANT}}}/oauth2/v2.0/token",
        "method": aiohttp.hdrs.METH_POST,
        "use_cookies": False,
        "data": {
            "client_info": "1",
            "scope": (
                "https://gruenbeckb2c.onmicrosoft.com/iot/user_impersonation openid profile offline_access"  # noqa: E501
            ),
            "grant_type": "refresh_token",
            "refresh_token": f"{{{PARAM_NAME_REFRESH_TOKEN}}}",
            "client_id": "5a83cc16-ffb1-42e9-9859-9fbf07f36df8",
        },
        "json_data": False,
        "query_params": {},
        "headers": {
            "Host": "gruenbeckb2c.b2clogin.com",
            "x-client-SKU": "MSAL.iOS",
            "Accept": "application/json",
            "x-client-OS": "14.3",
            "x-app-name": "Grünbeck",
            "x-client-CPU": "64",
            "x-app-ver": "1.2.0",
            "Accept-Language": "de-de",
            "client-request-id": "F2929DED-2C9D-49F5-A0F4-31215427667C",
            "User-Agent": USER_AGENT_APP,
            "x-client-Ver": "0.8.0",
            "x-client-DM": "iPhone",
            "return-client-request-id": "true",
            "cache-control": "no-cache",
        },
    },
    "start_ws_negotiation": {
        "scheme": API_SCHEME,
        "host": API_HOST,
        "port": DEFAULT_PORTS.get(API_SCHEME),
        "path": "/api/realtime/negotiate",
        "method": aiohttp.hdrs.METH_GET,
        "use_cookies": False,
        "data": {},
        "json_data": False,
        "query_params": {},
        "headers": {
            "Content-Type": "text/plain;charset=UTF-8",
            "Origin": "file://",
            "Accept": "*/*",
            "User-Agent": USER_AGENT_APP,
            "Authorization": f"Bearer {{{PARAM_NAME_ACCESS_TOKEN}}}",
            "Accept-Language": "de-de",
            "cache-control": "no-cache",
            "X-Requested-With": "XMLHttpRequest",
        },
    },
    "get_ws_connection_id": {
        "scheme": API_WS_SCHEME_HTTP,
        "host": API_WS_HOST,
        "port": DEFAULT_PORTS.get(API_WS_SCHEME_HTTP),
        "path": "/client/negotiate",
        "method": aiohttp.hdrs.METH_POST,
        "use_cookies": False,
        "data": {},
        "json_data": False,
        "query_params": {"hub": "gruenbeck"},
        "headers": {
            "Content-Type": "text/plain;charset=UTF-8",
            "Origin": "file://",
            "Accept": "*/*",
            "User-Agent": USER_AGENT_APP,
            "Authorization": f"Bearer {{{PARAM_NAME_ACCESS_TOKEN}}}",
            "Accept-Language": "de-de",
            "X-Requested-With": "XMLHttpRequest",
        },
    },
    "get_devices": {
        "scheme": API_SCHEME,
        "host": API_HOST,
        "port": DEFAULT_PORTS.get(API_SCHEME),
        "path": "/api/devices",
        "method": aiohttp.hdrs.METH_GET,
        "use_cookies": False,
        "data": {},
        "json_data": False,
        "query_params": {
            "api-version": API_VERSION,
        },
        "headers": {
            "Host": "prod-eu-gruenbeck-api.azurewebsites.net",
            "Accept": "application/json, text/plain, */*",
            "User-Agent": USER_AGENT_APP,
            "Accept-Language": "de-de",
            "Authorization": f"Bearer {{{PARAM_NAME_ACCESS_TOKEN}}}",
            "cache-control": "no-cache",
        },
    },
    "get_device_infos_request": {
        "scheme": API_SCHEME,
        "host": API_HOST,
        "port": DEFAULT_PORTS.get(API_SCHEME),
        "path": f"/api/devices/{{{PARAM_NAME_DEVICE_ID}}}/{{{PARAM_NAME_ENDPOINT}}}",
        "method": aiohttp.hdrs.METH_GET,
        "use_cookies": False,
        "data": {},
        "json_data": False,
        "query_params": {
            "api-version": API_VERSION,
        },
        "headers": {
            "Host": "prod-eu-gruenbeck-api.azurewebsites.net",
            "Accept": "application/json, text/plain, */*",
            "User-Agent": USER_AGENT_APP,
            "Accept-Language": "de-de",
            "Authorization": f"Bearer {{{PARAM_NAME_ACCESS_TOKEN}}}",
            "cache-control": "no-cache",
        },
    },
    "enter_sd": {
        "scheme": API_SCHEME,
        "host": API_HOST,
        "port": DEFAULT_PORTS.get(API_SCHEME),
        "path": f"/api/devices/{{{PARAM_NAME_DEVICE_ID}}}/realtime/enter",
        "method": aiohttp.hdrs.METH_POST,
        "use_cookies": False,
        "data": {},
        "json_data": False,
        "query_params": {
            "api-version": API_VERSION,
        },
        "headers": {
            "Host": "prod-eu-gruenbeck-api.azurewebsites.net",
            "Accept": "application/json, text/plain, */*",
            "User-Agent": USER_AGENT_APP,
            "Accept-Language": "de-de",
            "Authorization": f"Bearer {{{PARAM_NAME_ACCESS_TOKEN}}}",
        },
    },
    "refresh_sd": {
        "scheme": API_SCHEME,
        "host": API_HOST,
        "port": DEFAULT_PORTS.get(API_SCHEME),
        "path": f"/api/devices/{{{PARAM_NAME_DEVICE_ID}}}/realtime/refresh",
        "method": aiohttp.hdrs.METH_POST,
        "use_cookies": False,
        "data": {},
        "json_data": False,
        "query_params": {
            "api-version": API_VERSION,
        },
        "headers": {
            "Host": "prod-eu-gruenbeck-api.azurewebsites.net",
            "Accept": "application/json, text/plain, */*",
            "User-Agent": USER_AGENT_APP,
            "Accept-Language": "de-de",
            "Authorization": f"Bearer {{{PARAM_NAME_ACCESS_TOKEN}}}",
        },
    },
    "leave_sd": {
        "scheme": API_SCHEME,
        "host": API_HOST,
        "port": DEFAULT_PORTS.get(API_SCHEME),
        "path": f"/api/devices/{{{PARAM_NAME_DEVICE_ID}}}/realtime/leave",
        "method": aiohttp.hdrs.METH_POST,
        "use_cookies": False,
        "data": {},
        "json_data": False,
        "query_params": {
            "api-version": API_VERSION,
        },
        "headers": {
            "Host": "prod-eu-gruenbeck-api.azurewebsites.net",
            "Accept": "application/json, text/plain, */*",
            "User-Agent": USER_AGENT_APP,
            "Accept-Language": "de-de",
            "Authorization": f"Bearer {{{PARAM_NAME_ACCESS_TOKEN}}}",
        },
    },
    "update_device_parameter": {
        "scheme": API_SCHEME,
        "host": API_HOST,
        "port": DEFAULT_PORTS.get(API_SCHEME),
        "path": f"/api/devices/{{{PARAM_NAME_DEVICE_ID}}}/parameters",
        "method": aiohttp.hdrs.METH_PATCH,
        "use_cookies": False,
        "data": {},
        "json_data": True,
        "query_params": {
            "api-version": API_VERSION,
        },
        "headers": {
            "Host": "prod-eu-gruenbeck-api.azurewebsites.net",
            "Content-Type": "application/json",
            "Accept": "application/json, text/plain, */*",
            "User-Agent": USER_AGENT_APP,
            "Accept-Language": "de-de",
            "Authorization": f"Bearer {{{PARAM_NAME_ACCESS_TOKEN}}}",
        },
    },
    "regenerate": {
        "scheme": API_SCHEME,
        "host": API_HOST,
        "port": DEFAULT_PORTS.get(API_SCHEME),
        "path": f"/api/devices/{{{PARAM_NAME_DEVICE_ID}}}/regenerate",
        "method": aiohttp.hdrs.METH_POST,
        "use_cookies": False,
        "data": {},
        "json_data": True,
        "query_params": {
            "api-version": API_VERSION,
        },
        "headers": {
            "Host": "prod-eu-gruenbeck-api.azurewebsites.net",
            "Content-Type": "application/json",
            "Accept": "application/json, text/plain, */*",
            "User-Agent": USER_AGENT_APP,
            "Accept-Language": "de-de",
            "Authorization": f"Bearer {{{PARAM_NAME_ACCESS_TOKEN}}}",
        },
    },
    "placeholder": {
        "scheme": LOGIN_SCHEME,
        "host": LOGIN_HOST,
        "port": DEFAULT_PORTS.get(LOGIN_SCHEME),
        "path": "/path",
        "method": aiohttp.hdrs.METH_GET,
        "use_cookies": False,
        "data": {},
        "json_data": False,
        "query_params": {},
        "headers": {},
    },
}

# Diagnostic
DIAGNOSTIC_REDACTED: Final = "**REDACTED**"
DIAGNOSTIC_REGEX: list[dict[str, Any]] = [
    {
        "regex": re.compile(r"%3d([A-Za-z0-9_\-\.]+)(%26|\")"),
        "index": 0,
    },
    {
        "regex": re.compile(
            r"(access_token|id_token|client_info|resource|refresh_token|id|serialNumber|accessToken|connectionId|pmailadress|pname|ptelnr)\":\s*\"([A-Za-z0-9_\-\.\/\s@+]+)\""  # noqa: E501
        ),
        "index": 1,
    },
    {
        "regex": re.compile(r"Bearer ([A-Za-z0-9_\-\.]+)"),
        "index": 0,
    },
]

# Mapping of some Parameter from API
PARAMETER_OPERATION_MODES: dict[int, str] = {
    1: "Eco",
    2: "Comfort",
    3: "Power",
    4: "Individual",
}

PARAMETER_OPERATION_MODES_INDIVIDUAL: Final = 4

PARAMETER_REGENERATION_MODES: dict[int, str] = {
    0: "Automatic",
    1: "Fixed",
}

PARAMETER_WATER_UNITS: dict[int, str] = {
    1: "°dH",
    2: "°fH",
    3: "°e",
    4: "mol/m³",
    5: "ppm",
}

PARAMETER_REGENERATION_STEP: dict[int, str] = {
    0: "Inactive",
    10: "Fill salt tank",
    20: "Salting",
    30: "Displacement",
    40: "Backwashing",
    50: "Backwashing",
    60: "Washing out",
}

PARAMETER_LANGUAGES: dict[int, str] = {
    1: "German",
    2: "English",
    3: "French",
    4: "Italian",
    5: "Dutch",
    6: "Spanish",
    7: "Russian",
    9: "Danish",
}

PARAMETER_LED_MODES: dict[int, str] = {
    0: "Deactivated",
    1: "Permanent lightning",
    2: "In case of failure",
    3: "In case of operation by user + failure",
    4: "In case of water treatment + operation by user + failure",
}



"""Module for pygruenbeck_cloud Exceptions."""

from __future__ import annotations


class PyGruenbeckCloudError(Exception):
    """Generic PyGruenbeck exception."""


class PyGruenbeckCloudResponseError(Exception):
    """PyGruenbeck API Response Error exception."""


class PyGruenbeckCloudResponseStatusError(Exception):
    """PyGruenbeck API Response status Error exception."""


class PyGruenbeckCloudUpdateParameterError(Exception):
    """PyGruenbeck API Response Error on Parameter Update exception."""


class PyGruenbeckCloudConnectionError(Exception):
    """PyGruenbeck API Connection Error exception."""


class PyGruenbeckCloudConnectionClosedError(Exception):
    """PyGruenbeck API Connection Closed exception."""


class PyGruenbeckCloudMissingAuthTokenError(Exception):
    """PyGruenbeck Missing Auth Token exception."""



"""Models for Gruenbeck Cloud library."""

from dataclasses import dataclass, field
import datetime
import logging
from typing import Any

from dataclasses_json import LetterCase, config as json_config, dataclass_json
from marshmallow import fields as mm_fields

from pygruenbeck_cloud.const import (
    API_WS_RESPONSE_TYPE_DATA,
    API_WS_RESPONSE_TYPE_DATA_TARGETS,
    API_WS_RESPONSE_TYPE_PING,
    LOGIN_REFRESH_TIME_BEFORE_EXPIRE,
)
from pygruenbeck_cloud.exceptions import PyGruenbeckCloudError


@dataclass
class GruenbeckAuthToken:
    """Object holding auth tokens for gruenbeck cloud."""

    access_token: str
    refresh_token: str
    not_before: datetime.datetime
    expires_on: datetime.datetime
    expires_in: int
    tenant: str

    def is_expired(self) -> bool:
        """Return if token is expired or not."""
        return datetime.datetime.now() >= (
            self.expires_on - LOGIN_REFRESH_TIME_BEFORE_EXPIRE
        )


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DeviceError:
    """Object holding Device Error Information."""

    is_resolved: bool
    message: str
    type: str
    date: datetime.datetime = field(
        metadata=json_config(
            encoder=lambda value: value.strftime("%Y-%m-%dT%H:%M:%S.%f"),
            # Object seems to be UTC, so we need to set correct timezone
            decoder=lambda value: datetime.datetime.strptime(
                value, "%Y-%m-%dT%H:%M:%S.%f"
            ).replace(tzinfo=datetime.UTC),
        ),
    )


@dataclass_json
@dataclass
class DailyUsageEntry:
    """Object holding daily usage data."""

    value: int
    date: datetime.date = field(
        metadata=json_config(
            encoder=lambda value: value.strftime("%Y-%m-%d"),
            decoder=lambda value: datetime.datetime.strptime(value, "%Y-%m-%d").date(),
            mm_field=mm_fields.Date(format="%Y-%m-%d"),
        ),
    )


@dataclass_json
@dataclass
class DeviceParameters:
    """Object holding Device Parameters."""

    # Daylight saving time
    dlst: bool | None = field(
        default=None,
        metadata=json_config(field_name="pdlstauto"),
    )

    # Audio signal on error
    buzzer: bool | None = field(
        default=None,
        metadata=json_config(field_name="pbuzzer"),
    )
    # Audio signal release from [hh:mm]
    buzzer_from: datetime.time | None = field(
        default=None,
        metadata=json_config(field_name="pbuzzfrom"),
    )
    # Audio signal release until [hh:mm]
    buzzer_to: datetime.time | None = field(
        default=None,
        metadata=json_config(field_name="pbuzzto"),
    )

    # Notifications
    push_notification: bool | None = field(
        default=None,
        metadata=json_config(field_name="pallowpushnotification"),
    )
    email_notification: bool | None = field(
        default=None,
        metadata=json_config(field_name="pallowemail"),
    )

    # Water settings
    water_hardness_unit: int | None = field(
        default=None,
        metadata=json_config(field_name="phunit"),
    )
    raw_water_hardness: int | None = field(
        default=None,
        metadata=json_config(field_name="prawhard"),
    )
    soft_water_hardness: int | None = field(
        default=None,
        metadata=json_config(field_name="psetsoft"),
    )

    # Working mode
    mode: int | None = field(
        default=None,
        metadata=json_config(field_name="pmode"),
    )
    mode_individual_monday: int | None = field(
        default=None,
        metadata=json_config(field_name="pmodemo"),
    )
    mode_individual_tuesday: int | None = field(
        default=None,
        metadata=json_config(field_name="pmodetu"),
    )
    mode_individual_wednesday: int | None = field(
        default=None,
        metadata=json_config(field_name="pmodewe"),
    )
    mode_individual_thursday: int | None = field(
        default=None,
        metadata=json_config(field_name="pmodeth"),
    )
    mode_individual_friday: int | None = field(
        default=None,
        metadata=json_config(field_name="pmodefr"),
    )
    mode_individual_saturday: int | None = field(
        default=None,
        metadata=json_config(field_name="pmodesa"),
    )
    mode_individual_sunday: int | None = field(
        default=None,
        metadata=json_config(field_name="pmodesu"),
    )

    # Regeneration mode
    regeneration_mode: int | None = field(
        default=None,
        metadata=json_config(field_name="pregmode"),
    )

    regeneration_time_monday_1: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregmo1",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_monday_2: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregmo2",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_monday_3: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregmo3",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_tuesday_1: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregtu1",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_tuesday_2: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregtu2",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_tuesday_3: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregtu3",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_wednesday_1: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregwe1",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_wednesday_2: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregwe2",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_wednesday_3: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregwe3",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_thursday_1: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregth1",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_thursday_2: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregth2",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_thursday_3: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregth3",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_friday_1: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregfr1",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_friday_2: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregfr2",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_friday_3: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregfr3",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )
    # regeneration_time_saturday_1: str | None = field(
    #     default=None,
    #     metadata=json_config(field_name="pregsa1"),
    # )  # datetime.time | None = None
    regeneration_time_saturday_1: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregsa1",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_saturday_2: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregsa2",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_saturday_3: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregsa3",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_sunday_1: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregsu1",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_sunday_2: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregsu2",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    regeneration_time_sunday_3: datetime.time | None = field(
        default=None,
        metadata=json_config(
            field_name="pregsu3",
            encoder=lambda value: (
                value.strftime("%H:%M")
                if value is not None and value != "--:--"
                else None
            ),
            decoder=lambda value: (
                datetime.datetime.strptime(value, "%H:%M").time()
                if value is not None and value != "--:--"
                else None
            ),
            mm_field=mm_fields.Date(format="%H:%M"),
        ),
    )

    # Maintenance information [days]
    maintenance_interval: int | None = field(
        default=None,
        metadata=json_config(field_name="pmaintint"),
    )
    # Installer information
    installer_name: str | None = field(
        default=None,
        metadata=json_config(field_name="pname"),
    )
    installer_phone: str | None = field(
        default=None,
        metadata=json_config(field_name="ptelnr"),
    )
    installer_email: str | None = field(
        default=None,
        metadata=json_config(field_name="pmailadress"),
    )

    # Get date/time automatically (NTP)
    ntp_sync: bool | None = field(
        default=None,
        metadata=json_config(field_name="pntpsync"),
    )

    # Function fault signal contact
    fault_signal_contact: bool | None = field(
        default=None,
        metadata=json_config(field_name="pcfcontact"),
    )

    # KNX connection
    knx: bool | None = field(
        default=None,
        metadata=json_config(field_name="pknx"),
    )

    # Monitoring of nominal flow
    nominal_flow_monitoring: bool | None = field(
        default=None,
        metadata=json_config(field_name="pmonflow"),
    )

    # Disinfection monitoring
    disinfection_monitoring: bool | None = field(
        default=None,
        metadata=json_config(field_name="pmondisinf"),
    )

    # Illuminated LED ring mode
    led_ring_mode: int | None = field(
        default=None,
        metadata=json_config(field_name="pled"),
    )
    # Illuminated LED ring flashes for pre-alarm salt supply
    led_ring_flash_on_signal: bool | None = field(
        default=None,
        metadata=json_config(field_name="pledatsaltpre"),
    )
    # LED ring Brightness [%]
    led_ring_brightness: int | None = field(
        default=None,
        metadata=json_config(field_name="pledbright"),
    )

    # Residual capacity limit value [%]
    residual_capacity_limit: int | None = field(
        default=None,
        metadata=json_config(field_name="prescaplimit"),
    )

    # Current setpoint [mA]
    current_setpoint: int | None = field(
        default=None,
        metadata=json_config(field_name="pcurrent"),
    )

    # Charge [mAmin]
    charge: int | None = field(
        default=None,
        metadata=json_config(field_name="pload"),
    )

    # Interval of forced regeneration [days]
    interval_forced_regeneration: int | None = field(
        default=None,
        metadata=json_config(field_name="pforcedregdist"),
    )

    # End frequency regeneration valve [Hz]
    end_frequency_regeneration_valve: int | None = field(
        default=None,
        metadata=json_config(field_name="pfreqregvalve"),
    )
    # End frequency regeneration valve 2 [Hz]
    end_frequency_regeneration_valve_2: int | None = field(
        default=None,
        metadata=json_config(field_name="pfreqregvalve2"),
    )

    # End frequency blending valve [Hz]
    end_frequency_blending_valve: int | None = field(
        default=None,
        metadata=json_config(field_name="pfreqblendvalve"),
    )

    # Treatment volume [m³]
    treatment_volume: int | None = field(
        default=None,
        metadata=json_config(field_name="pvolume"),
    )

    # Soft water meter pulse rate [l/Imp]
    soft_water_meter_pulse_rate: float | None = field(
        default=None,
        metadata=json_config(field_name="ppratesoftwater"),
    )

    # Blending water meter pulse rate [l/Imp]
    blending_water_meter_pulse_rate: float | None = field(
        default=None,
        metadata=json_config(field_name="pprateblending"),
    )

    # Regeneration water meter pulse rate [l/Imp]
    regeneration_water_meter_pulse_rate: float | None = field(
        default=None,
        metadata=json_config(field_name="pprateregwater"),
    )

    # Capacity figure Monday [m³x°dH]
    capacity_figure_monday: int | None = field(
        default=None,
        metadata=json_config(field_name="psetcapmo"),
    )
    # Capacity figure Tuesday [m³x°dH]
    capacity_figure_tuesday: int | None = field(
        default=None,
        metadata=json_config(field_name="psetcaptu"),
    )
    # Capacity figure Wednesday [m³x°dH]
    capacity_figure_wednesday: int | None = field(
        default=None,
        metadata=json_config(field_name="psetcapwe"),
    )
    # Capacity figure Thursday [m³x°dH]
    capacity_figure_thursday: int | None = field(
        default=None,
        metadata=json_config(field_name="psetcapth"),
    )
    # Capacity figure Friday [m³x°dH]
    capacity_figure_friday: int | None = field(
        default=None,
        metadata=json_config(field_name="psetcapfr"),
    )
    # Capacity figure Saturday [m³x°dH]
    capacity_figure_saturday: int | None = field(
        default=None,
        metadata=json_config(field_name="psetcapsa"),
    )
    # Capacity figure Sunday [m³x°dH]
    capacity_figure_sunday: int | None = field(
        default=None,
        metadata=json_config(field_name="psetcapsu"),
    )

    # Nominal flow rate [m³/h]
    nominal_flow_rate: float | None = field(
        default=None,
        metadata=json_config(field_name="pnomflow"),
    )

    # Regeneration monitoring time [min]
    regeneration_monitoring_time: int | None = field(
        default=None,
        metadata=json_config(field_name="pmonregmeter"),
    )
    # Salting monitoring time [min]
    salting_monitoring_time: int | None = field(
        default=None,
        metadata=json_config(field_name="pmonsalting"),
    )

    # Slow rinse [min]
    slow_rinse: float | None = field(
        default=None,
        metadata=json_config(field_name="prinsing"),
    )

    # Backwash [l]
    backwash: int | None = field(
        default=None,
        metadata=json_config(field_name="pbackwash"),
    )

    # Washing out [l]
    washing_out: int | None = field(
        default=None,
        metadata=json_config(field_name="pwashingout"),
    )

    # Minimum filling volume smallest cap [l]
    minimum_filling_volume_smallest_cap: float | None = field(
        default=None,
        metadata=json_config(field_name="pminvolmincap"),
    )
    # Maximum filling volume smallest cap [l]
    maximum_filling_volume_smallest_cap: float | None = field(
        default=None,
        metadata=json_config(field_name="pmaxvolmincap"),
    )
    # Minimum filling volume largest cap [l]
    minimum_filling_volume_largest_cap: float | None = field(
        default=None,
        metadata=json_config(field_name="pminvolmaxcap"),
    )
    # Maximum filling volume largest cap [l]
    maximum_filling_volume_largest_cap: float | None = field(
        default=None,
        metadata=json_config(field_name="pmaxvolmaxcap"),
    )

    # Longest switch-on time chlorine cell [min]
    longest_switch_on_time_chlorine_cell: int | None = field(
        default=None,
        metadata=json_config(field_name="pmaxdurdisinfect"),
    )

    # Maximum remaining time regeneration [min]
    maximum_remaining_time_regeneration: int | None = field(
        default=None,
        metadata=json_config(field_name="pmaxresdurreg"),
    )

    # Current language
    language: int | None = field(
        default=None,
        metadata=json_config(field_name="planguage"),
    )

    # Programmable output function
    programmable_output_function: int | None = field(
        default=None,
        metadata=json_config(field_name="pprogout"),
    )
    # Programmable input function
    programmable_input_function: int | None = field(
        default=None,
        metadata=json_config(field_name="pprogin"),
    )

    # Reaction to power failure > 5 min
    reaction_to_power_failure: int | None = field(
        default=None,
        metadata=json_config(field_name="ppowerfail"),
    )

    # Activate/deactivate chlorine cell
    chlorine_cell_mode: int | None = field(
        default=None,
        metadata=json_config(field_name="pmodedesinf"),
    )

    # Blending monitoring
    blending_monitoring: int | None = field(
        default=None,
        metadata=json_config(field_name="pmonblend"),
    )

    # System overloaded
    system_overloaded: int | None = field(
        default=None,
        metadata=json_config(field_name="poverload"),
    )
    # Unknown Parameter
    ppressurereg: int | None = field(
        default=None,
        metadata=json_config(field_name="ppressurereg"),
    )
    # pdate: "[yyyy.mm.dd] Current date",
    # pclearerrmem: "Delete error memory",
    # pclearcntwater: "Reset water meter",
    # pclearcntreg: "Reset regeneration counter",


@dataclass_json
@dataclass
class DeviceRealtimeInfo:
    """Object holding WebSocket realtime Information."""

    # Soft water exchanger 1 [l]
    soft_water_quantity: int | None = field(
        default=None,
        metadata=json_config(field_name="mcountwater1"),
    )
    # Soft water exchanger 2 [l]
    soft_water_quantity_2: int | None = field(
        default=None,
        metadata=json_config(field_name="mcountwater2"),
    )
    # Regeneration counter
    regeneration_counter: int | None = field(
        default=None,
        metadata=json_config(field_name="mcountreg"),
    )
    # Flow rate exchanger 1 [m³/h]
    current_flow_rate: float | None = field(
        default=None,
        metadata=json_config(field_name="mflow1"),
    )
    # Flow rate exchanger 2 [m³/h]
    current_flow_rate_2: float | None = field(
        default=None,
        metadata=json_config(field_name="mflow2"),
    )
    # Soft water Exchanger 1 [m³]
    remaining_capacity_volume: float | None = field(
        default=None,
        metadata=json_config(field_name="mrescapa1"),
    )
    # Soft water Exchanger 2 [m³]
    remaining_capacity_volume_2: float | None = field(
        default=None,
        metadata=json_config(field_name="mrescapa2"),
    )
    # Residual capacity Exchanger 1 [%]
    remaining_capacity_percentage: int | None = field(
        default=None,
        metadata=json_config(field_name="mresidcap1"),
    )
    # Residual capacity Exchanger 2 [%]
    remaining_capacity_percentage_2: int | None = field(
        default=None,
        metadata=json_config(field_name="mresidcap2"),
    )
    # Salt-reach [days]
    salt_range: int | None = field(
        default=None,
        metadata=json_config(field_name="msaltrange"),
    )
    # Salt consumption [kg]
    salt_consumption: float | None = field(
        default=None,
        metadata=json_config(field_name="msaltusage"),
    )
    # Perform maintenance in [days]
    next_service: int | None = field(
        default=None,
        metadata=json_config(field_name="mmaint"),
    )
    # @TODO - This two parameter provide information about running regeneration
    #   if both have value 0 when no regeneration is running.
    #   Current observations:
    #   mremregstep -> on regeneration start it seems to increase from 0 it seems to
    #       increase slowly up to 1,5 and then it changes with a big step to > 4400
    #   mregstatus -> when mremregstep is between 0 and 1,5, mremregstep has the
    #       value 10 as soon as mremregstep is > 4400 then it changes to 20
    #       on mremregstep ~= 300 it changes to 30 and when < 10 it changes to 40
    # Remaining amount / time of current regeneration step
    regeneration_remaining_time: float | None = field(
        default=None, metadata=json_config(field_name="mremregstep")
    )
    # Regeneration step
    regeneration_step: int | None = field(
        default=None,
        metadata=json_config(field_name="mregstatus"),
    )
    # Make-up water volume [l]
    make_up_water_volume: int | None = field(
        default=None,
        metadata=json_config(field_name="mcountwatertank"),
    )
    # During [min] int?
    # during_min: int | None = field(
    #     default=None,
    #     metadata=json_config(field_name="mflowexc"),
    # )
    # # : "[Min]", int?
    # make_up_water_volume: int | None = field(
    #     default=None,
    #     metadata=json_config(field_name="mflowexc2reg1"),
    # )
    # # : "[Min]", int?
    # make_up_water_volume: int | None = field(
    #     default=None,
    #     metadata=json_config(field_name="mflowexc1reg2"),
    # )

    # Adsorber exhausted percentage [%]
    exhausted_percentage: int | None = field(
        default=None,
        metadata=json_config(field_name="mlifeadsorb"),
    )
    # Actual value soft water hardness [°dh] - int?
    actual_value_soft_water_hardness: int | None = field(
        default=None,
        metadata=json_config(field_name="mhardsoftw"),
    )
    # Capacity figure [m³x°dH]
    capacity_figure: float | None = field(
        default=None,
        metadata=json_config(field_name="mcapacity"),
    )
    # Flow rate peak value [m³/h]
    flow_rate_peak_value: float | None = field(
        default=None,
        metadata=json_config(field_name="mflowmax"),
    )
    # Exchanger 1 peak value [m³/h] - float?
    exchanger_peak_value: float | None = field(
        default=None,
        metadata=json_config(field_name="mflowmax1reg2"),
    )
    # Exchanger 2 peak value [m³/h] - float?
    exchanger_peak_value_2: float | None = field(
        default=None,
        metadata=json_config(field_name="mflowmax2reg1"),
    )
    # Last regeneration Exchanger 1 [hh:mm]
    last_regeneration_exchanger: datetime.time | None = field(
        default=None,
        metadata=json_config(field_name="mendreg1"),
    )
    # Last regeneration Exchanger 2 [hh:mm]
    last_regeneration_exchanger_2: datetime.time | None = field(
        default=None,
        metadata=json_config(field_name="mendreg2"),
    )
    # # [%]?
    # percentage: int | None = field(
    #     default=None,
    #     metadata=json_config(field_name="mregpercent1"),
    # )
    # # 2 [%]?
    # percentage_2: int | None = field(
    #     default=None,
    #     metadata=json_config(field_name="mregpercent2"),
    # )
    # Regeneration flow rate Exchanger 1 [l/h] - int?
    regeneration_flow_rate_exchanger: int | None = field(
        default=None,
        metadata=json_config(field_name="mflowreg1"),
    )
    # Regeneration flow rate Exchanger 2 [l/h] - int?
    regeneration_flow_rate_exchanger_2: int | None = field(
        default=None,
        metadata=json_config(field_name="mflowreg2"),
    )
    # Blending flow rate [m³/h] - float?
    blending_flow_rate: float | None = field(
        default=None,
        metadata=json_config(field_name="mflowblend"),
    )
    # Step indication regeneration valve 1
    step_indication_regeneration_valve: int | None = field(
        default=None,
        metadata=json_config(field_name="mstep1"),
    )
    # Step indication regeneration valve 2
    step_indication_regeneration_valve_2: int | None = field(
        default=None,
        metadata=json_config(field_name="mstep2"),
    )
    # Current chlorine [mA]
    current_chlorine: int | None = field(
        default=None,
        metadata=json_config(field_name="mcurrent"),
    )
    # Adsorber remaining amount of water [m³] - float?
    remaining_amount_of_water: float | None = field(
        default=None,
        metadata=json_config(field_name="mreswatadmod"),
    )


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Device:
    """Object holding Device Information."""

    type: int
    has_error: bool
    id: str
    series: str
    serial_number: str
    name: str
    register: bool
    _next_regeneration_raw: datetime.datetime | None = field(
        default=None,
        metadata=json_config(
            field_name="nextRegeneration",
            encoder=lambda value: (
                value.strftime("%Y-%m-%dT%H:%M:%S") if value is not None else None
            ),
            decoder=lambda value: datetime.datetime.strptime(
                value, "%Y-%m-%dT%H:%M:%S"
            ),
        ),
    )
    time_zone: datetime.tzinfo | None = field(
        default=None,
        metadata=json_config(
            encoder=lambda value: value,
            decoder=lambda value: datetime.datetime.strptime(value, "%z").tzinfo,
        ),
    )
    # Start-up date
    startup: datetime.date | None = field(
        default=None,
        metadata=json_config(
            encoder=lambda value: (
                value.strftime("%Y-%m-%d") if value is not None else None
            ),
            decoder=lambda value: datetime.datetime.strptime(value, "%Y-%m-%d").date(),
            # mm_field=mm_fields.DateTime(format="%Y-%m-%d"),
        ),
    )
    last_service: datetime.date | None = field(
        default=None,
        metadata=json_config(
            encoder=lambda value: (
                value.strftime("%Y-%m-%d") if value is not None else None
            ),
            decoder=lambda value: datetime.datetime.strptime(value, "%Y-%m-%d").date(),
            # mm_field=mm_fields.DateTime(format="%Y-%m-%d"),
        ),
    )
    errors: list[DeviceError] | None = field(
        default=None,
        metadata=json_config(
            encoder=lambda value: value,
            decoder=lambda value: DeviceError.schema().load(value, many=True),  # type: ignore[attr-defined]  # noqa: E501  # pylint: disable=no-member
        ),
    )

    salt: list[DailyUsageEntry] | None = field(
        default=None,
        metadata=json_config(
            encoder=lambda value: value,
            decoder=lambda value: DailyUsageEntry.schema().load(value, many=True),  # type: ignore[attr-defined]  # noqa: E501  # pylint: disable=no-member
        ),
    )
    water: list[DailyUsageEntry] | None = field(
        default=None,
        metadata=json_config(
            encoder=lambda value: value,
            decoder=lambda value: DailyUsageEntry.schema().load(value, many=True),  # type: ignore[attr-defined]  # noqa: E501  # pylint: disable=no-member
        ),
    )
    hardware_version: str | None = None
    mode: int | None = None
    nominal_flow: float | None = None
    raw_water: float | None = None
    soft_water: float | None = None
    software_version: str | None = None
    unit: int | None = None

    # Values from WebSocket
    realtime: DeviceRealtimeInfo = field(default_factory=DeviceRealtimeInfo)

    # Device Parameter Values
    parameters: DeviceParameters = field(default_factory=DeviceParameters)

    # Logger instance
    logger: logging.Logger = logging.getLogger(__name__)

    def update_from_dict(self, data: dict) -> "Device":
        """Update current object from json dict."""
        return self.from_dict(self.to_dict() | data)  # type: ignore[attr-defined,no-any-return]  # noqa: E501  # pylint: disable=no-member

    def update_from_response(self, data: dict[str, Any]) -> "Device":
        """Update object with data from API response."""
        # If we got PING, do nothing
        if data.get("type") == API_WS_RESPONSE_TYPE_PING:
            return self

        # Parse Message Data
        if data.get("type") == API_WS_RESPONSE_TYPE_DATA:
            if not data.get("target") in API_WS_RESPONSE_TYPE_DATA_TARGETS:
                self.logger.debug(
                    "Got unknown target '%s' in response: %s", data.get("target"), data
                )
                return self

            message_arguments = data.get("arguments")
            if not message_arguments:
                self.logger.error("No arguments found in response: %s", data)
                return self

            for message in message_arguments:
                if message.get("id") != self.serial_number:
                    msg = (
                        f"Expected id value {self.serial_number}"
                        f" but got {message.get('id')}"
                    )
                    raise PyGruenbeckCloudError(msg)

                self.realtime = DeviceRealtimeInfo.from_dict(  # type: ignore[attr-defined]  # noqa: E501  # pylint: disable=no-member
                    self.realtime.to_dict() | message  # type: ignore[attr-defined]  # noqa: E501  # pylint: disable=no-member
                )
        # Got an unknown response type
        else:
            self.logger.debug(
                "Got response type '%s' which we don't can process: %s",
                data.get("type"),
                data,
            )

        return self

    @property
    def next_regeneration(self) -> datetime.datetime | None:
        """Return next regeneration value."""
        if self._next_regeneration_raw is None:
            return None

        return self._next_regeneration_raw.replace(tzinfo=self.time_zone)



"""pygruenbeck_cloud is a Python library to communicate with the Grünbeck Cloud based Water softeners."""  # noqa: E501

from __future__ import annotations

import base64
from collections import deque
from collections.abc import Callable
import dataclasses
from datetime import datetime
import hashlib
import json
from json import JSONDecodeError
import logging
import random
import socket
from types import TracebackType
from typing import Any

import aiohttp
from aiohttp import (
    ClientConnectionError,
    ClientConnectorError,
    ClientSession,
    ClientTimeout,
    ClientWebSocketResponse,
    ContentTypeError,
    ServerDisconnectedError,
    WSMsgType,
    WSServerHandshakeError,
)
from aiohttp.typedefs import StrOrURL
from yarl import URL

from .const import (
    API_GET_MG_INFOS_ENDPOINT,
    API_GET_MG_INFOS_ENDPOINT_PARAMETERS,
    API_GET_MG_INFOS_ENDPOINT_SALT_MEASUREMENTS,
    API_GET_MG_INFOS_ENDPOINT_WATER_MEASUREMENTS,
    API_WS_CLIENT_HEADER,
    API_WS_CLIENT_QUERY,
    API_WS_CLIENT_URL,
    API_WS_HOST,
    API_WS_INITIAL_MESSAGE,
    API_WS_REQUEST_TIMEOUT,
    API_WS_SCHEME_WS,
    DIAGNOSTIC_REDACTED,
    DIAGNOSTIC_REGEX,
    LOGIN_CODE_CHALLENGE_CHARS,
    PARAM_NAME_ACCESS_TOKEN,
    PARAM_NAME_CODE,
    PARAM_NAME_CODE_CHALLENGE,
    PARAM_NAME_CODE_VERIFIER,
    PARAM_NAME_CONNECTION_ID,
    PARAM_NAME_CSRF_TOKEN,
    PARAM_NAME_DEVICE_ID,
    PARAM_NAME_ENDPOINT,
    PARAM_NAME_PASSWORD,
    PARAM_NAME_POLICY,
    PARAM_NAME_REFRESH_TOKEN,
    PARAM_NAME_TENANT,
    PARAM_NAME_TRANS_ID,
    PARAM_NAME_USERNAME,
    WEB_REQUESTS,
)
from .exceptions import (
    PyGruenbeckCloudConnectionClosedError,
    PyGruenbeckCloudConnectionError,
    PyGruenbeckCloudError,
    PyGruenbeckCloudMissingAuthTokenError,
    PyGruenbeckCloudResponseError,
    PyGruenbeckCloudResponseStatusError,
    PyGruenbeckCloudUpdateParameterError,
)
from .models import DailyUsageEntry, Device, DeviceParameters, GruenbeckAuthToken

_LOGGER = logging.getLogger(__name__)


class PyGruenbeckCloud:
    """Class for communicate with the Grünbeck cloud."""

    session: ClientSession | None = None
    _close_session: bool = False
    _ws_session: ClientSession | None = None
    _ws_client: ClientWebSocketResponse | None = None
    _auth_token: GruenbeckAuthToken | None = None
    _device: Device | None = None
    logger: logging.Logger = logging.getLogger(__name__)

    def __init__(self, username: str, password: str) -> None:
        """Initialize PyGruenbeckCloud Class."""
        self._username = username
        self._password = password

        # Last responses
        self._response_log: deque = deque(maxlen=25)

    @staticmethod
    def _placeholder_to_values_dict(
        const: dict[str, str], values: dict[str, str]
    ) -> dict[str, str]:
        """Convert placeholder from dict to value in dict."""
        result = {}
        for key, value in const.items():
            result[key] = value.format(**values)

        return result

    @staticmethod
    def _placeholder_to_values_str(const: str, values: dict[str, str]) -> str:
        """Convert placeholder from str to values in dict."""
        return const.format(**values)

    @staticmethod
    def _extract_from_html_response(
        response: str, search_str: str, sep: str = ","
    ) -> str:
        """Retrieve str from HTML response."""
        start = response.index(search_str) + len(search_str) + 3
        end = response.index(sep, start) - 1

        return response[start:end]

    @staticmethod
    async def _get_code_challenge() -> list[str]:
        """Get Grünbeck Cloud API Code Challenge."""
        challenge_hash = ""
        result = ""

        while (
            challenge_hash == ""
            or "+" in challenge_hash
            or "/" in challenge_hash
            or "=" in challenge_hash
            or "+" in result
            or "/" in result
        ):
            result = "".join(
                random.choice(LOGIN_CODE_CHALLENGE_CHARS) for _ in range(64)
            )
            result = base64.b64encode(result.encode()).decode().rstrip("=")
            hash_object = hashlib.sha256(result.encode())
            challenge_hash = base64.b64encode(hash_object.digest()).decode()[:-1]

        return [result, challenge_hash]

    async def get_diagnostics(self) -> list[dict[str, Any]]:
        """Return a dict with diagnostic information for debugging purposes."""
        result = []

        def _filter(str_value: str) -> str:
            # Replace serial_number
            if (
                isinstance(self.device, Device)
                and self.device.serial_number is not None
            ):
                str_value = str_value.replace(
                    self.device.serial_number, DIAGNOSTIC_REDACTED
                )

            for regex in DIAGNOSTIC_REGEX:
                index = regex["index"]
                for match in regex["regex"].findall(str_value):
                    if isinstance(match, tuple):
                        str_value = str_value.replace(match[index], DIAGNOSTIC_REDACTED)
                    else:
                        str_value = str_value.replace(match, DIAGNOSTIC_REDACTED)

            return str_value

        # We need to remove data for privacy reasons!
        for entry in self._response_log:
            new_entry = {}
            for key, value in entry.items():
                if isinstance(value, str):
                    new_entry[key] = _filter(value)
                elif isinstance(value, dict):
                    new_entry[key] = {}  # type: ignore[assignment]
                    for sub_key, sub_value in value.items():
                        # Replace PARAM_NAME_USERNAME and PARAM_NAME_PASSWORD !!
                        if sub_key in (PARAM_NAME_USERNAME, PARAM_NAME_PASSWORD):
                            new_entry[key][sub_key] = DIAGNOSTIC_REDACTED  # type: ignore[index]  # noqa: E501
                        else:
                            new_entry[key][sub_key] = _filter(sub_value)  # type: ignore[index]  # noqa: E501
                else:
                    new_entry[key] = value

                # We encode responses to base64 to reduce size
                if key == "response":
                    new_entry[key] = base64.b64encode(new_entry[key].encode("utf-8"))  # type: ignore[assignment]  # noqa: E501

            result.append(new_entry)

        return result

    async def login(self) -> bool:
        """Login to Grünbeck Cloud."""
        code_verifier, code_challenge = await self._get_code_challenge()

        auth_data = await self._login_step1(code_challenge)

        if not await self._login_step2(auth_data):
            self.logger.error("Unable to login")
            return False

        code = await self._login_step3(auth_data)

        response = await self._login_step4(auth_data, code, code_verifier)

        self._auth_token = GruenbeckAuthToken(
            access_token=response["access_token"],
            refresh_token=response["refresh_token"],
            not_before=datetime.fromtimestamp(response["not_before"]),
            expires_on=datetime.fromtimestamp(response["expires_on"]),
            expires_in=response["expires_in"],
            tenant=auth_data["tenant"],
        )

        return True

    async def _login_step1(self, code_challenge: str) -> dict[str, str]:
        scheme = WEB_REQUESTS["login_step_1"]["scheme"]
        host = WEB_REQUESTS["login_step_1"]["host"]
        port = WEB_REQUESTS["login_step_1"]["port"]
        use_cookies = WEB_REQUESTS["login_step_1"]["use_cookies"]

        headers = WEB_REQUESTS["login_step_1"]["headers"]
        path = WEB_REQUESTS["login_step_1"]["path"]
        method = WEB_REQUESTS["login_step_1"]["method"]
        data = WEB_REQUESTS["login_step_1"]["data"]

        query = self._placeholder_to_values_dict(
            WEB_REQUESTS["login_step_1"]["query_params"],
            {PARAM_NAME_CODE_CHALLENGE: code_challenge},
        )

        # If we already have cookies, we will get a 302 and our code_challenge will not
        # match, that's why we need to clear our cookies
        if self.session and self.session.cookie_jar:
            self.session.cookie_jar.clear()

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        response = await self._http_request(
            url=url,
            headers=headers,
            method=method,
            data=data,
            use_cookies=use_cookies,
        )
        if not isinstance(response, str):
            msg = f"Incorrect response from {url}"
            raise PyGruenbeckCloudResponseError(msg)

        return {
            "csrf_token": self._extract_from_html_response(
                response=response, search_str="csrf"
            ),
            "transId": self._extract_from_html_response(
                response=response, search_str="transId"
            ),
            "policy": self._extract_from_html_response(
                response=response, search_str="policy"
            ),
            "tenant": self._extract_from_html_response(
                response=response, search_str="tenant"
            ),
        }

    async def _login_step2(self, auth_data: dict[str, str]) -> bool:
        scheme = WEB_REQUESTS["login_step_2"]["scheme"]
        host = WEB_REQUESTS["login_step_2"]["host"]
        port = WEB_REQUESTS["login_step_2"]["port"]
        use_cookies = WEB_REQUESTS["login_step_2"]["use_cookies"]

        headers = self._placeholder_to_values_dict(
            WEB_REQUESTS["login_step_2"]["headers"],
            {PARAM_NAME_CSRF_TOKEN: auth_data["csrf_token"]},
        )

        path = self._placeholder_to_values_str(
            WEB_REQUESTS["login_step_2"]["path"],
            {PARAM_NAME_TENANT: auth_data["tenant"]},
        )

        data = self._placeholder_to_values_dict(
            WEB_REQUESTS["login_step_2"]["data"],
            {
                PARAM_NAME_USERNAME: self._username,
                PARAM_NAME_PASSWORD: self._password,
            },
        )

        method = WEB_REQUESTS["login_step_2"]["method"]

        query = self._placeholder_to_values_dict(
            WEB_REQUESTS["login_step_2"]["query_params"],
            {
                PARAM_NAME_TRANS_ID: auth_data["transId"],
                PARAM_NAME_POLICY: auth_data["policy"],
            },
        )

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        response = await self._http_request(
            url=url,
            headers=headers,
            method=method,
            data=data,
            use_cookies=use_cookies,
        )

        if "status" in response:
            parsed_response = {}
            if isinstance(response, str):
                parsed_response = json.loads(response)
            elif isinstance(response, dict):
                parsed_response = response
            else:
                msg = f"Incorrect response from {url}"
                raise PyGruenbeckCloudResponseError(msg)

            if parsed_response["status"] == "200":
                return True

        return False

    async def _login_step3(self, auth_data: dict[str, str]) -> str:
        scheme = WEB_REQUESTS["login_step_3"]["scheme"]
        host = WEB_REQUESTS["login_step_3"]["host"]
        port = WEB_REQUESTS["login_step_3"]["port"]
        use_cookies = WEB_REQUESTS["login_step_3"]["use_cookies"]

        headers = WEB_REQUESTS["login_step_3"]["headers"]
        path = self._placeholder_to_values_str(
            WEB_REQUESTS["login_step_3"]["path"],
            {PARAM_NAME_TENANT: auth_data["tenant"]},
        )
        method = WEB_REQUESTS["login_step_3"]["method"]
        data = WEB_REQUESTS["login_step_3"]["data"]

        query = self._placeholder_to_values_dict(
            WEB_REQUESTS["login_step_3"]["query_params"],
            {
                PARAM_NAME_CSRF_TOKEN: auth_data["csrf_token"],
                PARAM_NAME_TRANS_ID: auth_data["transId"],
                PARAM_NAME_POLICY: auth_data["policy"],
            },
        )

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        # @TODO - expected_status_codes and allow_redirects can also come from CONST!
        response = await self._http_request(
            url=url,
            headers=headers,
            method=method,
            data=data,
            expected_status_codes=[aiohttp.http.HTTPStatus.FOUND],
            allow_redirects=False,
            use_cookies=use_cookies,
        )

        if not isinstance(response, str):
            msg = f"Incorrect response from {url}"
            raise PyGruenbeckCloudResponseError(msg)

        start = response.index("code%3d") + 7
        end = response.index(">here") - 1
        return response[start:end]

    async def _login_step4(
        self, auth_data: dict[str, str], code: str, code_verifier: str
    ) -> dict[str, Any]:
        scheme = WEB_REQUESTS["login_step_4"]["scheme"]
        host = WEB_REQUESTS["login_step_4"]["host"]
        port = WEB_REQUESTS["login_step_4"]["port"]
        use_cookies = WEB_REQUESTS["login_step_4"]["use_cookies"]

        headers = WEB_REQUESTS["login_step_4"]["headers"]
        path = self._placeholder_to_values_str(
            WEB_REQUESTS["login_step_4"]["path"],
            {PARAM_NAME_TENANT: auth_data["tenant"]},
        )
        method = WEB_REQUESTS["login_step_4"]["method"]
        data = self._placeholder_to_values_dict(
            WEB_REQUESTS["login_step_4"]["data"],
            {PARAM_NAME_CODE: code, PARAM_NAME_CODE_VERIFIER: code_verifier},
        )
        query = WEB_REQUESTS["login_step_4"]["query_params"]

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        response = await self._http_request(
            url=url,
            headers=headers,
            method=method,
            data=data,
            use_cookies=use_cookies,
        )

        if not isinstance(response, dict):
            msg = f"Incorrect response from {url}"
            raise PyGruenbeckCloudResponseError(msg)

        return response

    async def _refresh_web_token(self) -> bool:
        """Refresh Web Access Token."""
        if not isinstance(self._auth_token, GruenbeckAuthToken):
            msg = "Cannot refresh, missing Auth Token."
            raise PyGruenbeckCloudMissingAuthTokenError(msg)

        scheme = WEB_REQUESTS["web_token_refresh"]["scheme"]
        host = WEB_REQUESTS["web_token_refresh"]["host"]
        port = WEB_REQUESTS["web_token_refresh"]["port"]
        use_cookies = WEB_REQUESTS["web_token_refresh"]["use_cookies"]

        headers = WEB_REQUESTS["web_token_refresh"]["headers"]
        path = self._placeholder_to_values_str(
            WEB_REQUESTS["web_token_refresh"]["path"],
            {PARAM_NAME_TENANT: self._auth_token.tenant},
        )
        method = WEB_REQUESTS["web_token_refresh"]["method"]
        data = self._placeholder_to_values_dict(
            WEB_REQUESTS["web_token_refresh"]["data"],
            {
                PARAM_NAME_REFRESH_TOKEN: self._auth_token.refresh_token,
            },
        )
        query = WEB_REQUESTS["web_token_refresh"]["query_params"]

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        response = await self._http_request(
            url=url,
            headers=headers,
            method=method,
            data=data,
            use_cookies=use_cookies,
        )
        if not isinstance(response, dict):
            msg = f"Incorrect response from {url}"
            raise PyGruenbeckCloudResponseError(msg)

        # @TODO - Check response if token is expired!

        self._auth_token.access_token = response["access_token"]
        self._auth_token.refresh_token = response["refresh_token"]
        self._auth_token.not_before = datetime.fromtimestamp(response["not_before"])
        self._auth_token.expires_on = datetime.fromtimestamp(response["expires_on"])
        self._auth_token.expires_in = response["expires_in"]

        return True

    async def get_devices(self) -> list[Device]:
        """Get Devices from Cloud."""
        devices: list[Device] = []

        token = await self._get_web_access_token()

        scheme = WEB_REQUESTS["get_devices"]["scheme"]
        host = WEB_REQUESTS["get_devices"]["host"]
        port = WEB_REQUESTS["get_devices"]["port"]
        use_cookies = WEB_REQUESTS["get_devices"]["use_cookies"]

        headers = self._placeholder_to_values_dict(
            WEB_REQUESTS["get_devices"]["headers"],
            {
                PARAM_NAME_ACCESS_TOKEN: token,
            },
        )
        path = WEB_REQUESTS["get_devices"]["path"]
        method = WEB_REQUESTS["get_devices"]["method"]
        data = WEB_REQUESTS["get_devices"]["data"]
        query = WEB_REQUESTS["get_devices"]["query_params"]

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        response = await self._http_request(
            url=url,
            headers=headers,
            method=method,
            data=data,
            use_cookies=use_cookies,
        )

        if not isinstance(response, list):
            msg = f"Incorrect response from {url}"
            raise PyGruenbeckCloudResponseError(msg)

        for device in response:
            if "soft" in device["id"]:
                devices.append(Device.from_dict(device))  # type: ignore[attr-defined]  # noqa: E501  # pylint: disable=no-member

        return devices

    @property
    def device(self) -> Device | None:
        """Return current device."""
        return self._device

    async def set_device(self, device: Device, init: bool = True) -> None:
        """Async setter for device."""
        self._device = device

        if init is True:
            # Set logger for Device
            self._device.logger = self.logger
            try:
                await self.get_device_infos()
            except PyGruenbeckCloudResponseError as ex:
                msg = "Unable to get device infos"
                raise PyGruenbeckCloudError(msg) from ex

    async def set_device_from_id(self, device_id: str) -> bool:
        """Set device from given device ID."""
        devices = await self.get_devices()

        for device in devices:
            if device.id == device_id:
                await self.set_device(device)
                return True

        return False

    async def get_device_infos(self) -> Device:
        """Retrieve information for device."""
        if self.device is None:
            msg = "You need to select a device first"
            raise PyGruenbeckCloudError(msg)

        data = await self._get_device_infos_request(
            self.device, API_GET_MG_INFOS_ENDPOINT
        )
        if not isinstance(data, dict):
            msg = "Incorrect response for get_device_infos"
            raise PyGruenbeckCloudResponseError(msg)

        if data.get("id") != self.device.id:
            msg = f"Got invalid device id {data.get('id')}, expected {self.device.id}"
            raise PyGruenbeckCloudResponseError(msg)

        # Update current device object
        await self.set_device(self.device.update_from_dict(data), init=False)

        return self.device

    async def get_device_infos_parameters(self) -> Device:
        """Retrieve parameters for device."""
        if self.device is None:
            msg = "You need to select a device first"
            raise PyGruenbeckCloudError(msg)

        data = await self._get_device_infos_request(
            self.device, API_GET_MG_INFOS_ENDPOINT_PARAMETERS
        )
        if not isinstance(data, dict):
            msg = "Incorrect response for get_device_infos_parameters"
            raise PyGruenbeckCloudResponseError(msg)

        self.device.parameters = DeviceParameters.from_dict(data)  # type: ignore[attr-defined]  # noqa: E501  # pylint: disable=no-member

        return self.device

    async def get_device_salt_measurements(self) -> Device:
        """Retrieve salt measurements for device."""
        if self.device is None:
            msg = "You need to select a device first"
            raise PyGruenbeckCloudError(msg)

        data = await self._get_device_infos_request(
            self.device, API_GET_MG_INFOS_ENDPOINT_SALT_MEASUREMENTS
        )
        if not isinstance(data, list):
            msg = "Incorrect response for get_device_salt_measurements"
            raise PyGruenbeckCloudResponseError(msg)

        self.device.salt = DailyUsageEntry.schema().load(data, many=True)  # type: ignore[attr-defined]  # noqa: E501  # pylint: disable=no-member

        return self.device

    async def get_device_water_measurements(self) -> Device:
        """Retrieve water measurements for device."""
        if self.device is None:
            msg = "You need to select a device first"
            raise PyGruenbeckCloudError(msg)

        data = await self._get_device_infos_request(
            self.device, API_GET_MG_INFOS_ENDPOINT_WATER_MEASUREMENTS
        )
        if not isinstance(data, list):
            msg = "Incorrect response for get_device_water_measurements"
            raise PyGruenbeckCloudResponseError(msg)

        self.device.water = DailyUsageEntry.schema().load(data, many=True)  # type: ignore[attr-defined]  # noqa: E501  # pylint: disable=no-member

        return self.device

    async def _get_device_infos_request(
        self, device: Device, endpoint: str = ""
    ) -> Any:
        """Get Device Infos from API."""
        token = await self._get_web_access_token()

        scheme = WEB_REQUESTS["get_device_infos_request"]["scheme"]
        host = WEB_REQUESTS["get_device_infos_request"]["host"]
        port = WEB_REQUESTS["get_device_infos_request"]["port"]
        use_cookies = WEB_REQUESTS["get_device_infos_request"]["use_cookies"]

        headers = self._placeholder_to_values_dict(
            WEB_REQUESTS["get_device_infos_request"]["headers"],
            {
                PARAM_NAME_ACCESS_TOKEN: token,
            },
        )
        path = self._placeholder_to_values_str(
            WEB_REQUESTS["get_device_infos_request"]["path"],
            {
                PARAM_NAME_DEVICE_ID: device.id,
                PARAM_NAME_ENDPOINT: endpoint,
            },
        )
        method = WEB_REQUESTS["get_device_infos_request"]["method"]
        data = WEB_REQUESTS["get_device_infos_request"]["data"]
        query = WEB_REQUESTS["get_device_infos_request"]["query_params"]

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        response = await self._http_request(
            url=url,
            headers=headers,
            method=method,
            data=data,
            use_cookies=use_cookies,
        )

        return response

    async def update_device_infos_parameters(
        self,
        data: dict[str, Any],
    ) -> Device:
        """Update device parameters."""
        if self.device is None:
            msg = "You need to select a device first"
            raise PyGruenbeckCloudError(msg)

        # We need a copy to have only changed parameter
        parameters = dataclasses.replace(self.device.parameters)
        for key, value in data.items():
            if hasattr(parameters, key):
                new_value = value
                # JSON must contain the right data type
                if not isinstance(getattr(parameters, key), type(new_value)):
                    data_type = type(getattr(parameters, key))
                    if data_type == int:
                        new_value = int(value)
                    elif data_type == float:
                        new_value = float(value)
                    elif data_type == bool:
                        new_value = bool(value)
                    elif data_type == datetime.time:
                        new_value = datetime.strptime(data[key], "%H:%M").time()
                setattr(parameters, key, new_value)

        # Get difference with right JSON Keys
        parameter_json_dict = parameters.to_dict()  # type: ignore[attr-defined]
        changed_data = dict(
            set(parameter_json_dict.items())
            ^ set(self.device.parameters.to_dict().items())  # type: ignore[attr-defined]  # noqa: E501
        )

        if not changed_data:
            self.logger.warning("No changes detected in provided parameters")
            return self.device

        # Create JSON Object with changed data
        json_data = {}
        for key in changed_data.keys():
            json_data[key] = parameter_json_dict[key]

        token = await self._get_web_access_token()

        scheme = WEB_REQUESTS["update_device_parameter"]["scheme"]
        host = WEB_REQUESTS["update_device_parameter"]["host"]
        port = WEB_REQUESTS["update_device_parameter"]["port"]
        use_cookies = WEB_REQUESTS["update_device_parameter"]["use_cookies"]

        headers = self._placeholder_to_values_dict(
            WEB_REQUESTS["update_device_parameter"]["headers"],
            {
                PARAM_NAME_ACCESS_TOKEN: token,
            },
        )
        path = self._placeholder_to_values_str(
            WEB_REQUESTS["update_device_parameter"]["path"],
            {
                PARAM_NAME_DEVICE_ID: self.device.id,
            },
        )
        method = WEB_REQUESTS["update_device_parameter"]["method"]
        query = WEB_REQUESTS["update_device_parameter"]["query_params"]

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        response = await self._http_request(
            url=url,
            headers=headers,
            method=method,
            json_data=json_data,
            use_cookies=use_cookies,
            expected_status_codes=[
                aiohttp.http.HTTPStatus.OK,
                aiohttp.http.HTTPStatus.INTERNAL_SERVER_ERROR,
            ],
        )
        if not isinstance(response, dict):
            msg = f"Incorrect response from {url}"
            raise PyGruenbeckCloudResponseError(msg)

        if "error" in response:
            error = (
                f"Unable to update parameter {list(data)}, "
                f"error from API is '{response['error']['type']}'."
            )
            self.logger.error(error)

            raise PyGruenbeckCloudUpdateParameterError(error)

        # Update current device parameters
        self.device.parameters = DeviceParameters.from_dict(response)  # type: ignore[attr-defined]  # noqa: E501  # pylint: disable=no-member

        return self.device

    async def regenerate(
        self,
    ) -> Device:
        """Start regeneration."""
        if self.device is None:
            msg = "You need to select a device first"
            raise PyGruenbeckCloudError(msg)

        token = await self._get_web_access_token()

        scheme = WEB_REQUESTS["regenerate"]["scheme"]
        host = WEB_REQUESTS["regenerate"]["host"]
        port = WEB_REQUESTS["regenerate"]["port"]
        use_cookies = WEB_REQUESTS["regenerate"]["use_cookies"]

        headers = self._placeholder_to_values_dict(
            WEB_REQUESTS["regenerate"]["headers"],
            {
                PARAM_NAME_ACCESS_TOKEN: token,
            },
        )
        path = self._placeholder_to_values_str(
            WEB_REQUESTS["regenerate"]["path"],
            {
                PARAM_NAME_DEVICE_ID: self.device.id,
            },
        )
        method = WEB_REQUESTS["regenerate"]["method"]
        data = WEB_REQUESTS["regenerate"]["data"]
        query = WEB_REQUESTS["regenerate"]["query_params"]

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        await self._http_request(
            url=url,
            headers=headers,
            method=method,
            json_data=data,
            use_cookies=use_cookies,
            expected_status_codes=[aiohttp.http.HTTPStatus.ACCEPTED],
        )

        return self.device

    async def enter_sd(
        self,
    ) -> None:
        """Send enter SD for WS."""
        if self.device is None:
            msg = "You need to select an device first"
            raise PyGruenbeckCloudError(msg)

        device = self.device

        token = await self._get_web_access_token()

        scheme = WEB_REQUESTS["enter_sd"]["scheme"]
        host = WEB_REQUESTS["enter_sd"]["host"]
        port = WEB_REQUESTS["enter_sd"]["port"]
        use_cookies = WEB_REQUESTS["enter_sd"]["use_cookies"]

        headers = self._placeholder_to_values_dict(
            WEB_REQUESTS["enter_sd"]["headers"],
            {
                PARAM_NAME_ACCESS_TOKEN: token,
            },
        )
        path = self._placeholder_to_values_str(
            WEB_REQUESTS["enter_sd"]["path"],
            {PARAM_NAME_DEVICE_ID: device.id},
        )
        method = WEB_REQUESTS["enter_sd"]["method"]
        data = WEB_REQUESTS["enter_sd"]["data"]
        query = WEB_REQUESTS["enter_sd"]["query_params"]

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        # @TODO - expected_status_codes and allow_redirects can also come from CONST!
        await self._http_request(
            url=url,
            headers=headers,
            method=method,
            data=data,
            expected_status_codes=[aiohttp.http.HTTPStatus.ACCEPTED],
            use_cookies=use_cookies,
        )

    async def refresh_sd(self) -> None:
        """Send refresh SD for WS."""
        if self.device is None:
            msg = "You need to select an device first"
            raise PyGruenbeckCloudError(msg)

        device = self.device

        token = await self._get_web_access_token()

        scheme = WEB_REQUESTS["refresh_sd"]["scheme"]
        host = WEB_REQUESTS["refresh_sd"]["host"]
        port = WEB_REQUESTS["refresh_sd"]["port"]
        use_cookies = WEB_REQUESTS["refresh_sd"]["use_cookies"]

        headers = self._placeholder_to_values_dict(
            WEB_REQUESTS["refresh_sd"]["headers"],
            {
                PARAM_NAME_ACCESS_TOKEN: token,
            },
        )
        path = self._placeholder_to_values_str(
            WEB_REQUESTS["refresh_sd"]["path"],
            {PARAM_NAME_DEVICE_ID: device.id},
        )
        method = WEB_REQUESTS["refresh_sd"]["method"]
        data = WEB_REQUESTS["refresh_sd"]["data"]
        query = WEB_REQUESTS["refresh_sd"]["query_params"]

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        # @TODO - expected_status_codes and allow_redirects can also come from CONST!
        await self._http_request(
            url=url,
            headers=headers,
            method=method,
            data=data,
            expected_status_codes=[aiohttp.http.HTTPStatus.ACCEPTED],
            use_cookies=use_cookies,
        )

    async def leave_sd(self) -> None:
        """Send leave SD for WS."""
        if self.device is None:
            msg = "You need to select an device first"
            raise PyGruenbeckCloudError(msg)

        device = self.device

        token = await self._get_web_access_token()

        scheme = WEB_REQUESTS["leave_sd"]["scheme"]
        host = WEB_REQUESTS["leave_sd"]["host"]
        port = WEB_REQUESTS["leave_sd"]["port"]
        use_cookies = WEB_REQUESTS["leave_sd"]["use_cookies"]

        headers = self._placeholder_to_values_dict(
            WEB_REQUESTS["leave_sd"]["headers"],
            {
                PARAM_NAME_ACCESS_TOKEN: token,
            },
        )
        path = self._placeholder_to_values_str(
            WEB_REQUESTS["leave_sd"]["path"],
            {PARAM_NAME_DEVICE_ID: device.id},
        )
        method = WEB_REQUESTS["leave_sd"]["method"]
        data = WEB_REQUESTS["leave_sd"]["data"]
        query = WEB_REQUESTS["leave_sd"]["query_params"]

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        # @TODO - expected_status_codes and allow_redirects can also come from CONST!
        await self._http_request(
            url=url,
            headers=headers,
            method=method,
            data=data,
            expected_status_codes=[aiohttp.http.HTTPStatus.ACCEPTED],
            use_cookies=use_cookies,
        )

    async def _http_request(
        self,
        *,
        headers: dict,
        url: StrOrURL,
        data: Any = None,
        json_data: Any = None,
        expected_status_codes: list[int] | None = None,
        method: str = aiohttp.hdrs.METH_GET,
        allow_redirects: bool = False,
        use_cookies: bool = False,
    ) -> str | dict[Any, Any] | list[Any]:
        """Execute HTTP request."""
        if expected_status_codes is None:
            expected_status_codes = [aiohttp.http.HTTPStatus.OK]

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            self.logger.debug("Requesting URL %s with method %s", url, method)
            async with self.session.request(
                method=method,
                url=url,
                headers=headers,
                allow_redirects=allow_redirects,
                data=data,
                json=json_data,
            ) as resp:
                try:
                    response = await resp.json()
                except ContentTypeError:
                    response = await resp.text()

                self.logger.debug(
                    "Response from URL %s with status %d was %s",
                    url,
                    resp.status,
                    response,
                )

                self._response_log.append(
                    {
                        "url": str(url),
                        "req_method": method,
                        "req_headers": headers,
                        "req_data": data,
                        "req_json_data": json_data,
                        "resp_headers": resp.headers,
                        "resp_status": str(resp.status),
                        "response": str(await resp.text()),
                    }
                )

                if (
                    not isinstance(response, str)
                    and not isinstance(response, dict)
                    and not isinstance(response, list)
                ):
                    msg = f"Response from URL {url} has incorrect type {type(response)}"
                    raise PyGruenbeckCloudResponseError(msg)

                if resp.status not in expected_status_codes:
                    error = (
                        f"Response status code for {url} is {resp.status},"
                        f" we expected one of {expected_status_codes}."
                    )
                    self.logger.error(error)

                    raise PyGruenbeckCloudResponseStatusError(error)

                if use_cookies:
                    self.session.cookie_jar.update_cookies(resp.cookies)

                return response
        except (ClientConnectorError, ServerDisconnectedError) as ex:
            self.logger.error("%s", ex)
            raise PyGruenbeckCloudConnectionError(ex) from ex

    @property
    def connected(self) -> bool:
        """Return if we are connected to WebSocket."""
        return self._ws_client is not None and not self._ws_client.closed

    async def connect(self) -> None:
        """Connect to the WebSocket."""
        if self.connected:
            return

        ws_access_token, ws_connection_id = await self._get_ws_tokens()

        query = self._placeholder_to_values_dict(
            API_WS_CLIENT_QUERY,
            {
                PARAM_NAME_CONNECTION_ID: ws_connection_id,
                PARAM_NAME_ACCESS_TOKEN: ws_access_token,
            },
        )
        url = URL.build(
            scheme=API_WS_SCHEME_WS,
            host=API_WS_HOST,
            path=API_WS_CLIENT_URL,
            query=query,
        )

        if self._ws_session is None:
            self._ws_session = ClientSession(
                timeout=ClientTimeout(total=API_WS_REQUEST_TIMEOUT)
            )

        try:
            self._ws_client = await self._ws_session.ws_connect(
                url=url, headers=API_WS_CLIENT_HEADER, heartbeat=30
            )
            # Send initial Message
            await self._ws_client.send_str(API_WS_INITIAL_MESSAGE)
            await self.enter_sd()
            await self.refresh_sd()
        except (
            WSServerHandshakeError,
            ClientConnectionError,
            socket.gaierror,
        ) as ex:
            raise PyGruenbeckCloudConnectionError(ex) from ex

    async def listen(self, callback: Callable[[Device], None]) -> None:
        """Listen for WebSocket messages."""
        if not self._ws_client or not self.connected:
            msg = "We are not connected to WebSocket"
            raise PyGruenbeckCloudConnectionError(msg)

        while not self._ws_client.closed:
            ws_msg = await self._ws_client.receive()
            self.logger.debug("WebSocket Message received: %s", ws_msg.data)
            self._response_log.append(
                {
                    "ws_tye": str(ws_msg.type),
                    "ws_data": str(
                        ws_msg.data,
                    ),
                }
            )

            if ws_msg.type == WSMsgType.ERROR:
                raise PyGruenbeckCloudConnectionError(self._ws_client.exception())

            if ws_msg.type == WSMsgType.TEXT:
                try:
                    # There is a "%1E = Record Separator" char at the end of the string!
                    response = json.loads(ws_msg.data.strip())

                    if response:
                        device = self.device.update_from_response(data=response)  # type: ignore[union-attr]  # noqa: E501
                        callback(device)
                    else:
                        self.logger.debug("Skipping empty response: %s", response)
                except JSONDecodeError:
                    self.logger.debug(
                        "Skipping invalid JSON response: %s", ws_msg.data.strip()
                    )

            if ws_msg.type == WSMsgType.BINARY:
                msg = "WebSocket response is binary type"
                raise PyGruenbeckCloudResponseError(msg)

            if ws_msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSED, WSMsgType.CLOSING):
                msg = "WebSocket connection has been closed"
                raise PyGruenbeckCloudConnectionClosedError(msg)

    async def disconnect(self) -> None:
        """Close open connections."""
        if not self._ws_session or not self.connected:
            return

        await self.leave_sd()
        await self._ws_session.close()

    async def _get_ws_tokens(self) -> list[str]:
        """Get new WebSocket tokens."""
        web_access_token = await self._get_web_access_token()

        _, ws_access_token = await self._start_ws_negotiation(
            access_token=web_access_token
        )
        ws_connection_id = await self._get_ws_connection_id(
            ws_access_token=ws_access_token
        )

        return [ws_access_token, ws_connection_id]

    async def _start_ws_negotiation(self, access_token: str) -> list[str]:
        """Start WebSocket connection negotiation."""
        scheme = WEB_REQUESTS["start_ws_negotiation"]["scheme"]
        host = WEB_REQUESTS["start_ws_negotiation"]["host"]
        port = WEB_REQUESTS["start_ws_negotiation"]["port"]
        use_cookies = WEB_REQUESTS["start_ws_negotiation"]["use_cookies"]

        headers = self._placeholder_to_values_dict(
            WEB_REQUESTS["start_ws_negotiation"]["headers"],
            {PARAM_NAME_ACCESS_TOKEN: access_token},
        )
        path = WEB_REQUESTS["start_ws_negotiation"]["path"]
        method = WEB_REQUESTS["start_ws_negotiation"]["method"]
        data = WEB_REQUESTS["start_ws_negotiation"]["data"]

        query = WEB_REQUESTS["start_ws_negotiation"]["query_params"]

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        response = await self._http_request(
            url=url,
            headers=headers,
            method=method,
            data=data,
            use_cookies=use_cookies,
        )

        if not isinstance(response, dict):
            msg = f"Incorrect response from {url}"
            raise PyGruenbeckCloudResponseError(msg)

        return [response["url"], response["accessToken"]]

    async def _get_ws_connection_id(self, ws_access_token: str) -> str:
        """Get WebSocket Connection ID."""
        scheme = WEB_REQUESTS["get_ws_connection_id"]["scheme"]
        host = WEB_REQUESTS["get_ws_connection_id"]["host"]
        port = WEB_REQUESTS["get_ws_connection_id"]["port"]
        use_cookies = WEB_REQUESTS["get_ws_connection_id"]["use_cookies"]

        headers = self._placeholder_to_values_dict(
            WEB_REQUESTS["get_ws_connection_id"]["headers"],
            {PARAM_NAME_ACCESS_TOKEN: ws_access_token},
        )
        path = WEB_REQUESTS["get_ws_connection_id"]["path"]
        method = WEB_REQUESTS["get_ws_connection_id"]["method"]
        data = WEB_REQUESTS["get_ws_connection_id"]["data"]

        query = WEB_REQUESTS["get_ws_connection_id"]["query_params"]

        url = URL.build(scheme=scheme, host=host, port=port, path=path, query=query)
        response = await self._http_request(
            url=url,
            headers=headers,
            method=method,
            data=data,
            use_cookies=use_cookies,
        )

        if not isinstance(response, dict):
            msg = f"Incorrect response from {url}"
            raise PyGruenbeckCloudResponseError(msg)

        return response["connectionId"]  # type: ignore[no-any-return]

    async def _get_web_access_token(self) -> str:
        """Get current WebSocket token."""
        if not isinstance(self._auth_token, GruenbeckAuthToken):
            await self.login()

        # Refreshes the token if needed
        if not self._auth_token.is_expired():  # type: ignore[union-attr]
            return self._auth_token.access_token  # type: ignore[union-attr]

        refresh = await self._refresh_web_token()
        if not refresh:
            self.logger.info("Unable to refresh token, need to relogin.")
            await self.login()

        return self._auth_token.access_token  # type: ignore[union-attr]

    async def close(self) -> None:
        """Close all connections."""
        await self.disconnect()

        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> PyGruenbeckCloud:
        """Start PyGruenbeckCloud class from context manager."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Stop PyGruenbeckCloud class from context manager."""
        await self.close()



"""pygruenbeck_cloud is a Python library to communicate with the Gruenbeck cloud."""

# flake8: noqa
from .pygruenbeck_cloud import PyGruenbeckCloud

__version__ = "1.1.0"  # pragma: no cover

__all__ = [
    "PyGruenbeckCloud",
]
